'use strict';

describe('epixa-resource', function() {
  beforeEach(module('epixa-resource'));

  var $rootScope, $httpBackend;
  beforeEach(inject(function($injector) {
    $rootScope = $injector.get('$rootScope');
    $httpBackend = $injector.get('$httpBackend');

    $httpBackend.whenGET('/foo').respond(200, [{id:1,foo:'bar'},{id:2,foo:'notbar'}]);

    $httpBackend.whenGET('/foo/1').respond({ id:1, 'foo':'bar' });
    $httpBackend.whenGET('/404').respond(404);

    $httpBackend.whenPOST('/foo').respond(201, {id:2, 'foo':'notbar'});
    $httpBackend.whenPOST('/with/pathfinder').respond(201, {id:1, 'something':'else'});
    $httpBackend.whenPOST('/500').respond(500);

    $httpBackend.whenGET('/already-stored').respond(200, {id:1, 'foo':'notbar'});
    $httpBackend.whenPUT('/already-stored').respond(200, {id:1, 'foo':'bar'});
    $httpBackend.whenPUT('/not-yet-stored').respond(200, {id:1, 'foo':'bar'});
    $httpBackend.whenPUT('/422').respond(422);

    $httpBackend.whenGET('/thing-to-delete').respond(200);
    $httpBackend.whenDELETE('/thing-to-delete').respond(202);
    $httpBackend.whenDELETE('/422').respond(422);

    $httpBackend.whenGET('/with/path/transformers').respond(200);
    $httpBackend.whenPOST('/with/path/transformers').respond(200, { id: 1 });
    $httpBackend.whenPUT('/with/path/transformers').respond(200);
    $httpBackend.whenDELETE('/with/path/transformers').respond(200);
  }));

  describe('resource-api', function() {
    var api;
    beforeEach(inject(function($injector) {
      api = $injector.get('resource-api');
    }));

    describe('.query()', function() {
      var collection, storedResource, initSpy, pathfinderSpy;
      beforeEach(function() {
        initSpy = jasmine.createSpy('initializer');
        pathfinderSpy = jasmine.createSpy('pathfinder').andCallFake(function(collectionPath, entity) {
          return collectionPath + '/' + entity.id;
        });
        storedResource = api.get('/foo/1'); // make sure it is stored in cache already
        $httpBackend.flush();
        resolveAll();
        collection = api.query('/foo', { initializer: initSpy, pathfinder: pathfinderSpy });
      });
      afterEach(function() {
        $httpBackend.verifyNoOutstandingExpectation();
      });
      describe('returned collection', function() {
        describe('.$path', function() {
          it('is equivalent to given path', function() {
            expect(collection.$path).toBe('/foo');
          });
        });
      });
      describe('when that path has not been seen before', function() {
        it('fires off a GET request to that path', function() {
          $httpBackend.expectGET('/foo');
        });
        describe('returned collection', function() {
          describe('when GET request is successful', function() {
            beforeEach(function() {
              $httpBackend.flush();
            });
            it('is populated with resources from response data', function() {
              expect(collection.resources.length).toBe(2);
            });
            describe('.$promise', function() {
              it('is resolved with collection', function() {
                expect(getResolvedValue(collection.$promise)).toBe(collection);
              });
            });
          });
          describe('when GET request fails', function() {
            beforeEach(function() {
              collection = api.query('/404');
              $httpBackend.flush();
            });
            describe('.$promise', function() {
              it('is rejected with http error', function() {
                expect(getRejectedValue(collection.$promise)).toBeHttpResponse();
              });
            });
          });
        });
        describe('when given an initializer function in the config (second argument)', function() {
          describe('initializer function', function() {
            it('is not called immediately', function() {
              expect(initSpy).not.toHaveBeenCalled();
            });
            describe('when .$promise resolves', function() {
              beforeEach(function() {
                $httpBackend.flush();
                resolveAll();
              });
              it('is called for each resource in the collection', function() {
                expect(initSpy.calls.length).toBe(2);
              });
              it('is passed the resource', function(){
                expect(initSpy).toHaveBeenCalledWith(collection.resources[0]);
                expect(initSpy).toHaveBeenCalledWith(collection.resources[1]);
              });
            });
          });
        });
        describe('when given a pathfinder function in the config (second argument)', function() {
          it('is not called immediately', function() {
            expect(pathfinderSpy).not.toHaveBeenCalled();
          });
          describe('when .$promise resolves', function() {
            beforeEach(function() {
              $httpBackend.flush();
              resolveAll();
            });
            it('is called for each resource in the collection', function() {
              expect(pathfinderSpy.calls.length).toBe(2);
            });
            it('is called with the original entity data and collection path for each resource', function(){
              expect(pathfinderSpy).toHaveBeenCalledWith('/foo', {id:1,foo:'bar'});
              expect(pathfinderSpy).toHaveBeenCalledWith('/foo', {id:2,foo:'notbar'});
            });
            describe('resource in collection', function() {
              describe('.$path', function() {
                it('is set to the result of the pathfinder', function() {
                  expect(collection.resources[0].$path).toBe('/foo/1');
                  expect(collection.resources[1].$path).toBe('/foo/2');
                });
              });
              describe('when already stored', function() {
                it('is identical to already stored resource (not just the same data)', function() {
                  expect(collection.resources[0]).toBe(storedResource);
                });
              });
              describe('when not already stored', function() {
                it('is stored for future requests', function() {
                  expect(collection.resources[1]).toBe(api.get('/foo/2'));
                });
              });
            });
          });
        });
        describe('when given an array of path transformers in the config (second argument)', function() {
          var collection, firstSpy, secondSpy;
          beforeEach(function() {
            firstSpy = jasmine.createSpy('pathTransformer1').andReturn('/notfoo');
            secondSpy = jasmine.createSpy('pathTransformer2').andReturn('/foo');
            collection = api.query('/transformers', { transformPath: [ firstSpy, secondSpy ] });
          });
          afterEach(function() {
            $httpBackend.verifyNoOutstandingExpectation();
          });
          it('passes the accumulative path to each transformer function', function() {
            expect(firstSpy).toHaveBeenCalledWith('/transformers');
            expect(secondSpy).toHaveBeenCalledWith('/notfoo');
          });
          it('decorates the http path in the array order', function() {
            $httpBackend.expectGET('/foo');
          });
          describe('returned collection', function() {
            describe('.$path', function() {
              it('is the same as original given path', function() {
                expect(collection.$path).toBe('/transformers');
              });
            });
          });
        });
      });
      describe('when that path has previously been seen', function() {
        var newCollection;
        beforeEach(function() {
          $httpBackend.flush();
          newCollection = api.query('/foo');
          resolveAll();
          collection.remove(collection.resources[0]);
          collection.add({$path:'/foo/not-1', foo:'bar'});
          collection.get('/foo/2').foo = 'somethingelse';
        });
        it('does not fire a GET request to that path', function() {
          $httpBackend.verifyNoOutstandingRequest();
        });
        it('returns the exact same collection object (not just the same data) as was previous seen', function() {
          expect(newCollection).toBe(collection);
        });
        describe('when given a reload flag in the config (second argument)', function() {
          beforeEach(function() {
            api.query('/foo', {reload: true});
          });
          it('fires off a GET request to that path', function() {
            $httpBackend.expectGET('/foo');
          });
          describe('stored collection', function() {
            describe('when GET request is successful', function() {
              beforeEach(function() {
                $httpBackend.flush();
                resolveAll();
              });
              it('has any new resources from http response', function() {
                var found = collection.resources.some(function(resource) {
                  return resource.$path === '/foo/1';
                });
                expect(found).toBe(true);
              });
              it('has no resources that were absent from the response', function() {
                var found = collection.resources.some(function(resource) {
                  return resource.$path === '/foo/not-1';
                });
                expect(found).toBe(false);
              });
              describe('resource in collection', function() {
                it('is extended with http response', function() {
                  expect(collection.get('/foo/2').foo).toBe('notbar');
                });
              });
            });
          });
        });
      });
    });

    describe('.get()', function() {
      var resource, initSpy;
      beforeEach(function() {
        initSpy = jasmine.createSpy('initializer');
        resource = api.get('/foo/1', { initializer: initSpy });
      });
      afterEach(function() {
        $httpBackend.verifyNoOutstandingExpectation();
      });
      describe('returned resource', function() {
        describe('.$path', function() {
          it('is equivalent to given path', function() {
            expect(resource.$path).toBe('/foo/1');
          });
        });
      });
      describe('when that path has not been seen before', function() {
        it('fires off a GET request to that path', function() {
          $httpBackend.expectGET('/foo/1');
        });
        describe('returned resource', function() {
          describe('when GET request is successful', function() {
            beforeEach(function() {
              $httpBackend.flush();
            });
            it('is extended with properties from response data', function() {
              expect(resource.foo).toBe('bar');
            });
            describe('.$promise', function() {
              it('is resolved with resource', function() {
                expect(getResolvedValue(resource.$promise)).toBe(resource);
              });
            });
          });
          describe('when GET request fails', function() {
            beforeEach(function() {
              resource = api.get('/404');
              $httpBackend.flush();
            });
            describe('.$promise', function() {
              it('is rejected with http error', function() {
                expect(getRejectedValue(resource.$promise)).toBeHttpResponse();
              });
            });
          });
        });
        describe('when given an initializer function in the config (second argument)', function() {
          describe('returned resource', function() {
            describe('when .$promise resolves', function() {
              beforeEach(function() {
                $httpBackend.flush();
                resolveAll();
              });
              it('is passed to the initializer function', function() {
                expect(initSpy).toHaveBeenCalledWith(resource);
              });
            });
          });
        });
        describe('when given an array of path transformers in the config (second argument)', function() {
          var resource, pathSpy, transformersSpy;
          beforeEach(function() {
            pathSpy = jasmine.createSpy('path').andReturn('/path/transformers');
            transformersSpy = jasmine.createSpy('transformers').andReturn('/with/path/transformers');
            resource = api.get('/transformers', { transformPath: [ pathSpy, transformersSpy ] });
          });
          afterEach(function() {
            $httpBackend.verifyNoOutstandingExpectation();
          });
          it('passes the accumulative path to each transformer function', function() {
            expect(pathSpy).toHaveBeenCalledWith('/transformers');
            expect(transformersSpy).toHaveBeenCalledWith('/path/transformers');
          });
          it('decorates the http path in the array order', function() {
            $httpBackend.expectGET('/with/path/transformers');
          });
          describe('returned resource', function() {
            describe('.$path', function() {
              describe('when http request completes', function() {
                beforeEach(function() {
                  $httpBackend.flush();
                  resolveAll();
                });
                it('is not affected', function() {
                  expect(resource.$path).toBe('/transformers');
                });
              });
            });
          });
        });
      });
      describe('when default path transformers are configured', function() {
        var pathSpy;
        beforeEach(function() {
          pathSpy = jasmine.createSpy('path').andReturn('/with/path/transformers');
          api.defaults.transformPath.push(pathSpy);
        });
        describe('default path transformers', function() {
          describe('when multiple http requests are made', function() {
            beforeEach(function() {
              api.get('/transformers', { reload: true });
              api.get('/transformers', { reload: true });
            });
            it('are only invoked once per request', function() {
              expect(pathSpy.calls.length).toBe(2);
            });
          });
        });
      });
      describe('when that path has previously been seen', function() {
        var newResource;
        beforeEach(function() {
          $httpBackend.flush();
          newResource = api.get('/foo/1');
          resolveAll();
        });
        it('does not fire a GET request to that path', function() {
          $httpBackend.verifyNoOutstandingRequest();
        });
        it('returns the exact same resource object (not just the same data) as was previous seen', function() {
          expect(newResource).toBe(resource);
        });
        describe('when given a reload flag in the config (second argument)', function() {
          beforeEach(function() {
            resource.foo = 'notbar';
            api.get('/foo/1', {reload: true});
          });
          it('fires off a GET request to that path', function() {
            $httpBackend.expectGET('/foo/1');
          });
          describe('stored resource', function() {
            describe('when GET request is successful', function() {
              beforeEach(function() {
                $httpBackend.flush();
              });
              it('is extended by http response body', function() {
                expect(resource.foo).toBe('bar');
              });
            });
          });
        });
      });
    });

    describe('.post()', function() {
      var resource, initSpy;
      beforeEach(function() {
        initSpy = jasmine.createSpy('initializer');
        resource = api.post('/foo', { foo: 'notbar' }, { initializer: initSpy });
      });
      afterEach(function() {
        $httpBackend.verifyNoOutstandingExpectation();
      });
      it('sends data object in POST request to the given path', function() {
        $httpBackend.expectPOST('/foo', { foo: 'notbar' });
      });
      describe('returned resource', function() {
        describe('.$path', function() {
          it('is null', function() {
            expect(resource.$path).toBe(null);
          });
          describe('when .$promise resolves', function() {
            beforeEach(function() {
              $httpBackend.flush();
              resolveAll();
            });
            it('is set to <given-path>/<resource.id>', function() {
              expect(resource.$path).toBe('/foo/2');
            });
          });
        });
        describe('when POST request fails', function() {
          beforeEach(function() {
            resource = api.post('/500');
            $httpBackend.flush();
          });
          describe('.$promise', function() {
            it('is rejected with http error', function() {
              expect(getRejectedValue(resource.$promise)).toBeHttpResponse();
            });
          });
        });
        describe('when POST request is successful', function() {
          beforeEach(function() {
            $httpBackend.flush();
            resolveAll();
          });
          it('is extended by POST request response body', function() {
            expect(resource.foo).toBe('notbar');
          });
          describe('.$promise', function() {
            it('is resolved with the resource', function() {
              expect(getResolvedValue(resource.$promise)).toBe(resource);
            });
          });
        });
        describe('when .$promise resolves', function() {
          beforeEach(function() {
            $httpBackend.flush();
            api.get('/foo/2');
            resolveAll();
          });
          it('is stored so subsequent calls to .get() for that resource do not fire a new request', function() {
            $httpBackend.verifyNoOutstandingRequest();
          });
        });
      });
      describe('when given an initializer function in the config (third argument)', function() {
        describe('returned resource', function() {
          describe('when .$promise resolves', function() {
            beforeEach(function() {
              $httpBackend.flush();
              resolveAll();
            });
            it('is passed to the initializer function', function() {
              expect(initSpy).toHaveBeenCalledWith(resource);
            });
          });
        });
      });
      describe('when given a custom pathfinder in the config (third argument)', function() {
        var pathfinder;
        beforeEach(function() {
          pathfinder = jasmine.createSpy('pathfinder').andReturn('/custom/pathfinder');
          resource = api.post('/with/pathfinder', { foo: 'notbar' }, {pathfinder: pathfinder});
        });
        describe('returned resource', function() {
          describe('.$path', function() {
            it('is null', function() {
              expect(resource.$path).toBe(null);
            });
            describe('when .$promise resolves', function() {
              beforeEach(function() {
                $httpBackend.flush();
                resolveAll();
              });
              it('is set to returned value from custom pathfinder', function() {
                expect(resource.$path).toBe('/custom/pathfinder');
              });
              describe('custom pathfinder', function() {
                it('is called with given path and returned resource', function() {
                  expect(pathfinder).toHaveBeenCalledWith('/with/pathfinder', resource);
                });
              });
            });
          });
        });
      });
      describe('when given an array of path transformers in the config (third argument)', function() {
        var resource, pathSpy, transformersSpy;
        beforeEach(function() {
          pathSpy = jasmine.createSpy('path').andReturn('/path/transformers');
          transformersSpy = jasmine.createSpy('transformers').andReturn('/with/path/transformers');
          resource = api.post('/transformers', {}, { transformPath: [ pathSpy, transformersSpy ] });
        });
        afterEach(function() {
          $httpBackend.verifyNoOutstandingExpectation();
        });
        it('passes the accumulative path to each transformer function', function() {
          expect(pathSpy).toHaveBeenCalledWith('/transformers');
          expect(transformersSpy).toHaveBeenCalledWith('/path/transformers');
        });
        it('decorates the http path in the array order', function() {
          $httpBackend.expectPOST('/with/path/transformers');
        });
        describe('returned resource', function() {
          describe('.$path', function() {
            describe('when http request completes', function() {
              beforeEach(function() {
                $httpBackend.flush();
                resolveAll();
              });
              it('is not affected', function() {
                expect(resource.$path).toBe('/transformers/1');
              });
            });
          });
        });
      });
    });

    describe('.put()', function() {
      beforeEach(function() {
        api.put('/already-stored', {foo: 'notbar'});
      });
      afterEach(function() {
        $httpBackend.verifyNoOutstandingExpectation();
      });
      it('sends PUT request to the given path', function() {
        $httpBackend.expectPUT('/already-stored', {foo: 'notbar'});
      });
      describe('returned promise', function() {
        describe('when PUT request fails', function() {
          var promise;
          beforeEach(function() {
            promise = api.put('/422');
            $httpBackend.flush();
          });
          it('is rejected with http error', function() {
            expect(getRejectedValue(promise)).toBeHttpResponse();
          });
        });
      });
      describe('when given a path that already has been stored', function() {
        var resource, promise;
        beforeEach(function() {
          resource = api.get('/already-stored');
          $httpBackend.flush(); // ensure it is stored
          promise = api.put('/already-stored', {foo: 'bar'});
          $httpBackend.flush();
        });
        describe('returned promise', function() {
          it('is fulfilled by the stored resource', function() {
            expect(getResolvedValue(promise)).toBe(resource);
          });
        });
        describe('stored resource', function() {
          it('is updated with http response', function() {
            expect(resource.foo).toBe('bar');
          });
        });
      });
      describe('when given a path that has not been stored', function() {
        var promise;
        beforeEach(function() {
          promise = api.put('/not-yet-stored', {foo: 'bar'});
          $httpBackend.flush();
        });
        describe('returned promise', function() {
          it('is fulfilled with resource', function() {
            expect(getResolvedValue(promise)).toBeResource();
          });
          describe('resolved resource', function() {
            var resource;
            beforeEach(function() {
              resource = getResolvedValue(promise);
              api.get('/not-yet-stored');
              resolveAll();
            });
            it('is stored for future reuqests for the same $path', function() {
              $httpBackend.verifyNoOutstandingRequest();
            });
            describe('.$path', function() {
              it('is set to given path', function() {
                expect(resource.$path).toBe('/not-yet-stored');
              });
              it('is populated with http response body', function() {
                expect(resource.id).toBe(1);
              });
            });
          });
        });
      });
      describe('when given an initializer function in the config (third argument)', function() {
        var promise, initSpy;
        beforeEach(function() {
          initSpy = jasmine.createSpy('initializer');
          promise = api.put('/not-yet-stored', {foo: 'bar'}, {initializer: initSpy});
          $httpBackend.flush();
        });
        describe('returned promise', function() {
          describe('resolved resource', function() {
            var resource;
            beforeEach(function() {
              resource = getResolvedValue(promise);
            });
            it('is passed to the initializer function', function() {
              expect(initSpy).toHaveBeenCalledWith(resource);
            });
          });
        });
      });
      describe('when given an array of path transformers in the config (third argument)', function() {
        var promise, pathSpy, transformersSpy;
        beforeEach(function() {
          pathSpy = jasmine.createSpy('path').andReturn('/path/transformers');
          transformersSpy = jasmine.createSpy('transformers').andReturn('/with/path/transformers');
          promise = api.put('/transformers', {}, { transformPath: [ pathSpy, transformersSpy ] });
        });
        afterEach(function() {
          $httpBackend.verifyNoOutstandingExpectation();
        });
        it('passes the accumulative path to each transformer function', function() {
          expect(pathSpy).toHaveBeenCalledWith('/transformers');
          expect(transformersSpy).toHaveBeenCalledWith('/path/transformers');
        });
        it('decorates the http path in the array order', function() {
          $httpBackend.expectPUT('/with/path/transformers');
        });
        describe('returned promise', function() {
          describe('resolved resource', function() {
            var resource;
            beforeEach(function() {
              $httpBackend.flush();
              resource = getResolvedValue(promise);
            });
            describe('.$path', function() {
              it('is not affected', function() {
                expect(resource.$path).toBe('/transformers');
              });
            });
          });
        });
      });
    });

    describe('.delete()', function() {
      beforeEach(function() {
        api.delete('/thing-to-delete');
      });
      afterEach(function() {
        $httpBackend.verifyNoOutstandingExpectation();
      });
      it('sends DELETE request to the given path', function() {
        $httpBackend.expectDELETE('/thing-to-delete');
      });
      describe('returned promise', function() {
        describe('when DELETE request fails', function() {
          var promise;
          beforeEach(function() {
            promise = api.delete('/422');
            $httpBackend.flush();
          });
          it('is rejected with http error', function() {
            expect(getRejectedValue(promise)).toBeHttpResponse();
          });
        });
        describe('when DELETE request is successful', function() {
          var promise;
          beforeEach(function() {
            promise = api.delete('/thing-to-delete');
            $httpBackend.flush();
          });
          it('is resolved with http response', function() {
            expect(getResolvedValue(promise)).toBeHttpResponse();
          });
        });
      });
      describe('when path matches the .$path of a stored resource', function() {
        var promise;
        beforeEach(function() {
          api.get('/thing-to-delete'); // ensure the thing is stored before delete
          promise = api.delete('/thing-to-delete');
          $httpBackend.flush();
        });
        describe('stored resource', function() {
          describe('when returned promise resolves', function() {
            beforeEach(function() {
              api.get('/thing-to-delete');
            });
            afterEach(function() {
              $httpBackend.verifyNoOutstandingExpectation();
            });
            it('is removed from cache so future requests to that resource do fire an http request', function() {
              $httpBackend.expectGET('/thing-to-delete');
            });
          });
        });
      });
      describe('when given an array of path transformers in the config (second argument)', function() {
        var pathSpy, transformersSpy;
        beforeEach(function() {
          pathSpy = jasmine.createSpy('path').andReturn('/path/transformers');
          transformersSpy = jasmine.createSpy('transformers').andReturn('/with/path/transformers');
          api.delete('/transformers', { transformPath: [ pathSpy, transformersSpy ] });
        });
        afterEach(function() {
          $httpBackend.verifyNoOutstandingExpectation();
        });
        it('passes the accumulative path to each transformer function', function() {
          expect(pathSpy).toHaveBeenCalledWith('/transformers');
          expect(transformersSpy).toHaveBeenCalledWith('/path/transformers');
        });
        it('decorates the http path in the array order', function() {
          $httpBackend.expectDELETE('/with/path/transformers');
        });
      });
    });
  });

  function resolveAll() {
    $rootScope.$apply(); // resolves angular promises
  };
  function getResolvedValue(promise) {
    var fulfilledBy = undefined;
    var resolved = false;
    promise.then(function(val) {
      resolved = true;
      fulfilledBy = val;
    });
    resolveAll();
    if (!resolved) throw new Error('Promise is not resolved at all');
    return fulfilledBy;
  }
  function getRejectedValue(promise) {
    var rejectedBy = undefined;
    var rejected = false;
    promise.then(null, function(val) {
      rejected = true;
      rejectedBy = val;
    });
    resolveAll();
    if (!rejected) throw new Error('Promise is not rejected at all');
    return rejectedBy;
  }
});
